package ru.dbotthepony.kstarbound.io

import it.unimi.dsi.fastutil.bytes.ByteArrayList
import ru.dbotthepony.kommons.core.IStruct2d
import ru.dbotthepony.kommons.core.IStruct2i
import ru.dbotthepony.kommons.io.readDouble
import ru.dbotthepony.kommons.io.readLong
import ru.dbotthepony.kommons.io.readSignedVarInt
import ru.dbotthepony.kommons.io.writeDouble
import ru.dbotthepony.kommons.io.writeLong
import ru.dbotthepony.kommons.io.writeSignedVarInt
import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kommons.vector.Vector2i
import ru.dbotthepony.kstarbound.world.ChunkPos
import java.io.DataInput
import java.io.DataOutput
import java.io.IOException
import java.io.InputStream
import java.io.OutputStream
import java.io.RandomAccessFile
import java.util.*
import java.util.function.IntConsumer
import java.util.function.IntSupplier
import kotlin.collections.ArrayList

fun RandomAccessFile.readString(length: Int): String {
	require(length >= 0) { "Invalid length $length" }

	val bytes = ByteArray(length)
	try {
		readFully(bytes)
	} catch(err: Throwable) {
		throw IOException("Tried to read string with length of $length", err)
	}

	return bytes.toString(Charsets.UTF_8)
}

fun InputStream.readString(length: Int, allowSmaller: Boolean = false): String {
	require(length >= 0) { "Invalid length $length" }

	val bytes = ByteArray(length)

	try {
		val read = readNBytes(bytes, 0, length)

		if (!allowSmaller)
			require(read == bytes.size) { "Read $read bytes, expected ${bytes.size}" }
	} catch(err: Throwable) {
		throw IOException("Tried to read string with length of $length", err)
	}

	return bytes.toString(Charsets.UTF_8)
}

fun InputStream.readByteChar(): Char {
	return read().toChar()
}

fun InputStream.readHeader(header: String) {
	for ((i, char) in header.withIndex()) {
		val read = readByteChar()

		if (read != char) {
			throw IllegalArgumentException("Malformed header at byte $i: expected $char (${char.code}) got $read (${char.code})")
		}
	}
}

fun OutputStream.writeUTF(value: String) {
	write(value.toByteArray().also { check(!it.any { it.toInt() == 0 }) { "Provided UTF string contains NUL" } })
	write(0)
}

fun InputStream.readUTF(): String {
	val bytes = ByteArrayList()
	var read = read()

	while (read != 0) {
		bytes.add(read.toByte())
		read = read()
	}

	return String(bytes.toByteArray())
}

fun InputStream.readUUID(): UUID {
	return UUID(readLong(), readLong())
}

fun OutputStream.writeUUID(value: UUID) {
	writeLong(value.mostSignificantBits)
	writeLong(value.leastSignificantBits)
}

fun OutputStream.writeVec2i(value: IStruct2i) {
	writeSignedVarInt(value.component1())
	writeSignedVarInt(value.component2())
}

fun OutputStream.writeVec2d(value: IStruct2d) {
	writeDouble(value.component1())
	writeDouble(value.component2())
}

fun InputStream.readVec2i(): Vector2i {
	return Vector2i(readSignedVarInt(), readSignedVarInt())
}

fun InputStream.readVec2d(): Vector2d {
	return Vector2d(readDouble(), readDouble())
}

fun InputStream.readChunkPos(): ChunkPos {
	return ChunkPos(readSignedVarInt(), readSignedVarInt())
}
